package com.jpattern.batch.core;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.atomic.AtomicBoolean;

import com.jpattern.batch.Job;
import com.jpattern.batch.JobExecutor;
import com.jpattern.batch.JobExecutionStrategy;
import com.jpattern.batch.JobPool;
import com.jpattern.batch.JobTrigger;
import com.jpattern.batch.execution.RunOnceJobExecutionStrategy;
import com.jpattern.batch.logger.JobPoolLogger;
import com.jpattern.batch.status.JobStatus;
import com.jpattern.logger.ILogger;

/**
 * 
 * @author Francesco Cina'
 *
 * 29/mar/2010
 */
public class ThreadJobPool implements JobPool {

	private static final long serialVersionUID = 1L;
	private final Map<String, Map<String, JobTrigger>> executableJobMap = new TreeMap<String, Map<String, JobTrigger>>();
	private boolean shutdown = false;
	private final AtomicBoolean serverStarted = new AtomicBoolean( false );
	private final String name;
	private final ILogger logger = JobPoolLogger.getLogger(this.getClass());

	public ThreadJobPool(final String applicationName) {
		name = applicationName;
		logger.info("JobPool [" + applicationName + "] created");
	}

	@Override
	public void addJob(final Job job, final JobExecutionStrategy jobExecutionStrategy) {
		logger.info("JobPool [" + name + "]: adding new job; name [" + job.getName() + "] group [" + job.getGroup() + "]");
		final JobExecutorImpl jobExecutor = new JobExecutorImpl(job);
		registerJob( jobExecutionStrategy, jobExecutor );
	}

	private void registerJob(final JobExecutionStrategy jobExecutionStrategy, final JobExecutor jobExecutor) {
		final String jobGroup = jobExecutor.getJobStatus().getGroupName();
		final String jobName = jobExecutor.getJobStatus().getJobName();
		if (!executableJobMap.containsKey(jobGroup)) {
			executableJobMap.put(jobGroup, new TreeMap<String, JobTrigger>());
		}
		final Map<String, JobTrigger> groupMap = executableJobMap.get(jobGroup);
		final CallableJobTrigger trigger = new CallableJobTrigger(jobExecutor, jobExecutionStrategy, serverStarted);
		groupMap.put(jobName, trigger);
		getExecutorService().submit( trigger );
	}

	@Override
	public JobStatus getJobStatus(final String jobName, final String groupName) {
		return getJobExecutor(jobName, groupName).getJobStatus();
	}

	@Override
	public List<String> getJobsName(final String groupName) {
		final List<String> result = new ArrayList<String>();
		if (executableJobMap.containsKey(groupName)) {
			final Map<String, JobTrigger> group = executableJobMap.get(groupName);
			for( final Entry<String, JobTrigger> jobEntry : group.entrySet() ) {
				result.add( jobEntry.getKey() );
			}
		}
		return result;
	}

	@Override
	public List<String> getGroupsName() {
		final List<String> result = new ArrayList<String>();
		for( final Entry<String, Map<String, JobTrigger>> jobEntry : executableJobMap.entrySet() ) {
			result.add( jobEntry.getKey() );
		}
		return result;
	}

	@Override
	public boolean isStarted() {
		return serverStarted.get();
	}

	@Override
	public void modifyJobExecution(final String jobName, final String groupName, final JobExecutionStrategy jobExecutionStrategy) {
		logger.info("JobPool [" + name + "]: rescheduling job name [" + jobName + "] group [" + groupName + "]");
		final JobExecutor executor = removeTrigger(jobName, groupName);
		if (executor.isValid()) {
			registerJob( jobExecutionStrategy, executor );
		}
	}

	@Override
	public void executeJob(final String jobName, final String groupName) {
		logger.info("JobPool [" + name + "]: execute job name [" + jobName + "] group [" + groupName + "]");
		final CallableJobTrigger trigger = new CallableJobTrigger(getJobExecutor(jobName, groupName), new RunOnceJobExecutionStrategy(), serverStarted);
		getExecutorService().submit(trigger);
	}

	@Override
	public void executeGroup(final String groupName) {
		logger.info("JobPool [" + name + "]: execute group [" + groupName + "]");
		for( final String jobName : getJobsName(groupName) ) {
			executeJob(jobName, groupName);
		}
	}

	@Override
	public void executeAllJobs() {
		logger.info("JobPool [" + name + "]: execute all jobs");
		for( final String groupName : getGroupsName() ) {
			executeGroup(groupName);
		}
	}

	@Override
	public void pauseJob(final String jobName, final String groupName) {
		logger.info("JobPool [" + name + "]: pause job name [" + jobName + "] group [" + groupName + "]");
		getJobExecutor(jobName, groupName).pause();
	}

	@Override
	public void pauseGroup(final String groupName) {
		logger.info("JobPool [" + name + "]: pause group [" + groupName + "]");
		for( final String jobName : getJobsName(groupName) ) {
			pauseJob(jobName, groupName);
		}
	}

	@Override
	public void pauseAllJobs() {
		logger.info("JobPool [" + name + "]: pause all jobs");
		for( final String groupName : getGroupsName() ) {
			pauseGroup(groupName);
		}
	}

	@Override
	public void resumeJob(final String jobName, final String groupName) {
		logger.info("JobPool [" + name + "]: resume job name [" + jobName + "] group [" + groupName + "]");
		getJobExecutor(jobName, groupName).resume();
	}

	@Override
	public void resumeGroup(final String groupName) {
		logger.info("JobPool [" + name + "]: resume group [" + groupName + "]");
		for( final String jobName : getJobsName(groupName) ) {
			resumeJob(jobName, groupName);
		}
	}

	@Override
	public void resumeAllJobs() {
		logger.info("JobPool [" + name + "]: resume all jobs");
		for( final String groupName : getGroupsName() ) {
			resumeGroup(groupName);
		}
	}

	@Override
	public void removeJob(final String jobName, final String groupName) {
		logger.info("JobPool [" + name + "]: remove job name [" + jobName + "] group [" + groupName + "]");
		removeTrigger(jobName, groupName).kill();
	}

	private JobExecutor removeTrigger(final String jobName, final String groupName) {
		if (executableJobMap.containsKey(groupName)) {
			final Map<String, JobTrigger> group = executableJobMap.get(groupName);
			if ( group.containsKey(jobName) ) {
				final JobTrigger trigger = group.get(jobName);
				final JobExecutor jobExecutor = trigger.getJobExecutor();
				group.remove(jobName);
				trigger.killTrigger();
				return jobExecutor;
			}
		}
		return new NullJobExecutor();

	}

	@Override
	public void removeGroup(final String groupName) {
		logger.info("JobPool [" + name + "]: remove group [" + groupName + "]");
		for( final String jobName : getJobsName(groupName) ) {
			removeJob(jobName, groupName);
		}
	}

	@Override
	public void removeAllJobs() {
		logger.info("JobPool [" + name + "]: remove all jobs");
		for( final String groupName : getGroupsName() ) {
			removeGroup(groupName);
		}
	}

	@Override
	public void shutdown(final boolean waitJobsExecutionEnd) {
		if (!shutdown){
			logger.info("JobPool [" + name +  "]: Begin engine shutdown... ");
			serverStarted.set(false);
			for( final String groupName : getGroupsName() ) {
				for( final String jobName : getJobsName(groupName) ) {
					getJobExecutor(jobName, groupName).kill();
				}
			}
			if (waitJobsExecutionEnd) {
				logger.info("JobPool [" + name +  "]: Waiting job executions end... ");
				boolean allExecutionEnded = false;
				while (!allExecutionEnded) {
					allExecutionEnded = true;
					for (final String groupName : getGroupsName() ) {
						for(final String jobName : getJobsName(groupName)) {
							if ( getJobStatus(jobName, groupName).isRunning()) {
								allExecutionEnded = false;
							}
						}
					}
					try {
						Thread.sleep(100);
					} catch (final InterruptedException e) {
						e.printStackTrace();
					}
				}
				logger.info("JobPool [" + name +  "]: Job executions ended ");
			}
			shutdown = true;
			logger.info("JobPool [" + name +  "]: Shutdown engine finished.");
		}
	}

	@Override
	public void start() {
		if (shutdown) {
			throw new RuntimeException("JobPool [" + name +  "]: Cannot start after shutdown");
		}
		if (!serverStarted.get()) {
			logger.info("JobPool [" + name +  "]: starting engine... ");
			serverStarted.set(true);
			logger.info("JobPool [" + name +  "]: engine started.");
		}
	}

	@Override
	public String getName() {
		return name;
	}

	@Override
	public boolean isShutdown() {
		return shutdown;
	}

	/**
	 * This should be changed with a pool to control the number of concurrent Threads
	 * @return
	 */
	private ExecutorService getExecutorService() {
		return Executors.newSingleThreadExecutor();
	}

	protected JobTrigger getJobTrigger(final String jobName, final String groupName) {
		if (executableJobMap.containsKey(groupName)) {
			final Map<String, JobTrigger> group = executableJobMap.get(groupName);
			if ( group.containsKey(jobName) ) {
				return group.get(jobName);
			}
		}
		return new NullJobTrigger();
	}

	private JobExecutor getJobExecutor(final String jobName, final String groupName) {
		return getJobTrigger(jobName, groupName).getJobExecutor();
	}
}
